import os
import sys
from pathlib import Path
from typing import Iterable
from unittest.mock import ANY

from pytest import fixture, raises

from getdents._getdents import DT_DIR, MIN_GETDENTS_BUFF_SIZE, getdents_raw


@fixture
def fixt_regular_file(tmp_path: Path) -> Iterable[int]:
    f = tmp_path / "test.txt"
    f.write_text("content")

    fd = os.open(f, os.O_RDONLY)

    yield fd

    os.close(fd)


@fixture
def fixt_dir(tmp_path: Path) -> Iterable[int]:
    for i in range(10):
        (tmp_path / f"subdir{i}").mkdir()

    fd = os.open(tmp_path, os.O_DIRECTORY | os.O_RDONLY)

    yield fd

    os.close(fd)


def test_not_a_directory(fixt_regular_file: int) -> None:
    with raises(NotADirectoryError):
        getdents_raw(fixt_regular_file, MIN_GETDENTS_BUFF_SIZE)


def test_small_buffer(fixt_dir: int) -> None:
    with raises(ValueError):
        getdents_raw(fixt_dir, MIN_GETDENTS_BUFF_SIZE - 1)


def test_malloc_fail(fixt_dir: int) -> None:
    with raises(MemoryError):
        getdents_raw(fixt_dir, sys.maxsize)


def test_getdents_raw(fixt_dir: int) -> None:
    iterator = iter(
        sorted(
            getdents_raw(
                fixt_dir,
                MIN_GETDENTS_BUFF_SIZE,
            ),
            key=lambda d: d[2],
        )
    )

    assert next(iterator) == (ANY, DT_DIR, ".")
    assert next(iterator) == (ANY, DT_DIR, "..")

    for i, entry in enumerate(iterator):
        assert entry == (ANY, DT_DIR, "subdir%d" % i)



from unittest.mock import Mock, patch, sentinel

from pytest import raises

from getdents import DT_DIR, DT_REG, DT_UNKNOWN, O_GETDENTS, getdents


@patch("getdents.os")
@patch("getdents.getdents_raw")
def test_path(mock_getdents_raw: Mock, mock_os: Mock) -> None:
    mock_os.open.return_value = sentinel.fd

    list(getdents("/tmp", sentinel.size))

    mock_os.open.assert_called_once_with("/tmp", O_GETDENTS)
    mock_getdents_raw.assert_called_once_with(sentinel.fd, sentinel.size)
    mock_os.close.assert_called_once_with(sentinel.fd)


@patch("getdents.os")
@patch("getdents.getdents_raw", side_effect=[Exception])
def test_path_err(mock_getdents_raw: Mock, mock_os: Mock) -> None:
    mock_os.open.return_value = sentinel.fd

    with raises(Exception):
        list(getdents("/tmp", sentinel.size))

    mock_os.open.assert_called_once_with("/tmp", O_GETDENTS)
    mock_getdents_raw.assert_called_once_with(sentinel.fd, sentinel.size)
    mock_os.close.assert_called_once_with(sentinel.fd)


@patch("getdents.os")
@patch(
    "getdents.getdents_raw",
    return_value=iter(
        [
            (1, DT_DIR, "."),
            (2, DT_DIR, ".."),
            (3, DT_DIR, "dir"),
            (4, DT_REG, "file"),
            (5, DT_UNKNOWN, "???"),
            (0, DT_REG, "deleted"),
        ]
    ),
)
def test_filtering(mock_getdents_raw: Mock, mock_os: Mock) -> None:
    assert list(getdents(".")) == [
        (3, DT_DIR, "dir"),
        (4, DT_REG, "file"),
    ]



from typing import Iterator, Tuple

DT_BLK: int
DT_CHR: int
DT_DIR: int
DT_FIFO: int
DT_LNK: int
DT_REG: int
DT_SOCK: int
DT_UNKNOWN: int
MIN_GETDENTS_BUFF_SIZE: int
O_GETDENTS: int

def getdents_raw(fd: int, buff_size: int) -> Iterator[Tuple[int, int, str]]: ...



from io import StringIO
from typing import Any, Iterable, Iterator, TextIO

from pytest import fixture

from getdents import DirectoryEntry
from getdents._getdents import (
    DT_BLK,
    DT_CHR,
    DT_DIR,
    DT_FIFO,
    DT_LNK,
    DT_REG,
    DT_SOCK,
    DT_UNKNOWN,
)
from getdents.formatters import (
    Echo,
    FormatterRegistry,
    format_csv,
    format_csv_headers,
    format_json,
    format_json_stream,
    format_plain,
    formatter,
    json_encode,
)


@fixture
def dirents() -> Iterator[DirectoryEntry]:
    return iter(
        [
            (1, DT_BLK, "block_device"),
            (2, DT_CHR, "character_device"),
            (3, DT_DIR, "directory"),
            (4, DT_FIFO, "named_pipe"),
            (5, DT_LNK, "symbolic_link"),
            (6, DT_REG, "file"),
            (7, DT_SOCK, "socket"),
            (8, DT_UNKNOWN, "unknown"),
        ]
    )


def test_formatter() -> None:
    def fn(
        directory_entries: Iterable[DirectoryEntry],
        file: TextIO,
    ) -> None:
        """Do nothing"""

    registry: FormatterRegistry = {}

    assert formatter("test", registry=registry)(fn) is fn
    assert registry == {"test": fn}


def test_format_plain(dirents: Iterator[DirectoryEntry]) -> None:
    output = StringIO()

    format_plain(dirents, output)

    assert output.getvalue() == (
        "block_device\n"
        "character_device\n"
        "directory\n"
        "named_pipe\n"
        "symbolic_link\n"
        "file\n"
        "socket\n"
        "unknown\n"
    )


def test_echo() -> None:
    echo = Echo()
    value: Any = object()

    assert echo.write(value) is value


def test_format_csv(dirents: Iterator[DirectoryEntry]) -> None:
    output = StringIO()

    format_csv(dirents, output)

    assert output.getvalue() == (
        "1,blk,block_device\r\n"
        "2,chr,character_device\r\n"
        "3,dir,directory\r\n"
        "4,fifo,named_pipe\r\n"
        "5,lnk,symbolic_link\r\n"
        "6,reg,file\r\n"
        "7,sock,socket\r\n"
        "8,unknown,unknown\r\n"
    )


def test_format_csv_headers(dirents: Iterator[DirectoryEntry]) -> None:
    output = StringIO()

    format_csv_headers(dirents, output)

    assert output.getvalue() == (
        "inode,type,name\r\n"
        "1,blk,block_device\r\n"
        "2,chr,character_device\r\n"
        "3,dir,directory\r\n"
        "4,fifo,named_pipe\r\n"
        "5,lnk,symbolic_link\r\n"
        "6,reg,file\r\n"
        "7,sock,socket\r\n"
        "8,unknown,unknown\r\n"
    )


def test_json_encode() -> None:
    assert json_encode(0, DT_UNKNOWN, "test") == (
        '{"inode": 0, "type": "unknown", "name": "test"}'
    )


def test_format_json(dirents: Iterator[DirectoryEntry]) -> None:
    output = StringIO()

    format_json(dirents, output)

    assert output.getvalue() == (
        "[\n"
        '{"inode": 1, "type": "blk", "name": "block_device"},\n'
        '{"inode": 2, "type": "chr", "name": "character_device"},\n'
        '{"inode": 3, "type": "dir", "name": "directory"},\n'
        '{"inode": 4, "type": "fifo", "name": "named_pipe"},\n'
        '{"inode": 5, "type": "lnk", "name": "symbolic_link"},\n'
        '{"inode": 6, "type": "reg", "name": "file"},\n'
        '{"inode": 7, "type": "sock", "name": "socket"},\n'
        '{"inode": 8, "type": "unknown", "name": "unknown"}\n'
        "]\n"
    )


def test_format_json_stream(dirents: Iterator[DirectoryEntry]) -> None:
    output = StringIO()

    format_json_stream(dirents, output)

    assert output.getvalue() == (
        '{"inode": 1, "type": "blk", "name": "block_device"}\n'
        '{"inode": 2, "type": "chr", "name": "character_device"}\n'
        '{"inode": 3, "type": "dir", "name": "directory"}\n'
        '{"inode": 4, "type": "fifo", "name": "named_pipe"}\n'
        '{"inode": 5, "type": "lnk", "name": "symbolic_link"}\n'
        '{"inode": 6, "type": "reg", "name": "file"}\n'
        '{"inode": 7, "type": "sock", "name": "socket"}\n'
        '{"inode": 8, "type": "unknown", "name": "unknown"}\n'
    )



from argparse import ArgumentParser
from sys import stderr, stdout
from typing import List, Optional, Tuple

from . import MIN_GETDENTS_BUFF_SIZE, getdents
from .formatters import FORMATTERS, Formatter


def parse_args(
    args: Optional[List[str]],
    prog: Optional[str],
) -> Tuple[str, int, Formatter]:
    parser = ArgumentParser(
        prog=prog,
        description="Print directory contents.",
    )

    parser.add_argument("path", metavar="PATH")
    parser.add_argument(
        "-b",
        "--buffer-size",
        metavar="N",
        type=int,
        default=32768,
        help=("Buffer size (in bytes) to allocate when iterating over directory"),
    )
    parser.add_argument(
        "-o",
        "--output-format",
        metavar="NAME",
        default="plain",
        choices=list(FORMATTERS),
        help="Output format: %s" % ", ".join(sorted(FORMATTERS)),
    )

    parsed_args = parser.parse_args(args)
    buff_size = parsed_args.buffer_size

    if buff_size < MIN_GETDENTS_BUFF_SIZE:
        parser.error("Minimum buffer size is %s" % MIN_GETDENTS_BUFF_SIZE)

    return parsed_args.path, buff_size, FORMATTERS[parsed_args.output_format]


def main(args: Optional[List[str]] = None, prog: Optional[str] = None) -> int:
    path, buff_size, fmt = parse_args(args, prog)

    try:
        fmt(getdents(path, buff_size=buff_size), stdout)
    except MemoryError:
        print(
            "Not enough memory to allocate",
            buff_size,
            "bytes of data",
            file=stderr,
        )
        return 3
    except FileNotFoundError as e:
        print(e, file=stderr)
        return 4
    except NotADirectoryError as e:
        print(e, file=stderr)
        return 5
    except PermissionError as e:
        print(e, file=stderr)
        return 6
    except OSError as e:
        print(e, file=stderr)
        return 7

    return 0



import os
from typing import Iterator, Tuple

from ._getdents import (
    DT_BLK,
    DT_CHR,
    DT_DIR,
    DT_FIFO,
    DT_LNK,
    DT_REG,
    DT_SOCK,
    DT_UNKNOWN,
    MIN_GETDENTS_BUFF_SIZE,
    O_GETDENTS,
    getdents_raw,
)

DirectoryEntry = Tuple[int, int, str]


def getdents(path: str, buff_size: int = 32768) -> Iterator[DirectoryEntry]:
    """Get directory entries.

    Wrapper around getdents_raw(), simulates ls behaviour: ignores deleted
    files, skips . and .. entries.

    Note:
       Default buffer size is 32k, it's a default allocation size of glibc's
       readdir() implementation.

    Note:
       Larger buffer will result in a fewer syscalls, so for really large
       dirs you should pick larger value.

    Note:
       For better performance, set buffer size to be multiple of your block
       size for filesystem I/O.

    Args:
        path (str): Location of the directory.
        buff_size (int): Buffer size in bytes for getdents64 syscall.
    """

    fd = os.open(path, O_GETDENTS)

    try:
        yield from (
            (inode, type, name)
            for inode, type, name in getdents_raw(fd, buff_size)
            if not (type == DT_UNKNOWN or inode == 0 or name in (".", ".."))
        )
    finally:
        os.close(fd)


__all__ = [
    "DT_BLK",
    "DT_CHR",
    "DT_DIR",
    "DT_FIFO",
    "DT_LNK",
    "DT_REG",
    "DT_SOCK",
    "DT_UNKNOWN",
    "MIN_GETDENTS_BUFF_SIZE",
    "O_GETDENTS",
    "DirectoryEntry",
    "getdents",
    "getdents_raw",
]



from csv import writer as csv_writer
from itertools import chain
from json import dumps as json_dumps
from sys import stdout
from typing import Callable, Dict, Iterable, TextIO

from . import (
    DT_BLK,
    DT_CHR,
    DT_DIR,
    DT_FIFO,
    DT_LNK,
    DT_REG,
    DT_SOCK,
    DT_UNKNOWN,
    DirectoryEntry,
)

Formatter = Callable[[Iterable[DirectoryEntry], TextIO], None]
FormatterRegistry = Dict[str, Formatter]
HEADER = ("inode", "type", "name")
FORMATTERS: FormatterRegistry = {}
TYPE_NAMES = {
    DT_BLK: "blk",
    DT_CHR: "chr",
    DT_DIR: "dir",
    DT_FIFO: "fifo",
    DT_LNK: "lnk",
    DT_REG: "reg",
    DT_SOCK: "sock",
    DT_UNKNOWN: "unknown",
}


def formatter(
    name: str,
    registry: Dict[str, Formatter] = FORMATTERS,
) -> Callable[[Formatter], Formatter]:
    def deco(fn: Formatter) -> Formatter:
        registry[name] = fn
        return fn

    return deco


@formatter("plain")
def format_plain(
    directory_entries: Iterable[DirectoryEntry],
    file: TextIO = stdout,
) -> None:
    for inode, type, name in directory_entries:
        print(name, file=file)


class Echo:
    def write(self, value: str) -> str:
        return value


def _format_csv(
    directory_entries: Iterable[DirectoryEntry],
    file: TextIO = stdout,
    headers: bool = False,
) -> None:
    writer = csv_writer(Echo())

    for first in directory_entries:
        if headers:
            print(writer.writerow(HEADER), end="", file=file)

        for inode, type, name in chain((first,), directory_entries):
            print(
                writer.writerow((inode, TYPE_NAMES[type], name)),
                end="",
                file=file,
            )


@formatter("csv")
def format_csv(
    directory_entries: Iterable[DirectoryEntry],
    file: TextIO = stdout,
) -> None:
    return _format_csv(directory_entries, file, False)


@formatter("csv-headers")
def format_csv_headers(
    directory_entries: Iterable[DirectoryEntry],
    file: TextIO = stdout,
) -> None:
    return _format_csv(directory_entries, file, True)


def json_encode(inode: int, type: int, name: str) -> str:
    return json_dumps(
        {
            "inode": inode,
            "type": TYPE_NAMES[type],
            "name": name,
        }
    )


@formatter("json")
def format_json(
    directory_entries: Iterable[DirectoryEntry],
    file: TextIO = stdout,
) -> None:
    for inode, type, name in directory_entries:
        print(
            "[\n",
            json_encode(inode, type, name),
            sep="",
            end="",
            file=file,
        )

        for inode, type, name in directory_entries:
            print(
                ",\n",
                json_encode(inode, type, name),
                sep="",
                end="",
                file=file,
            )

        print("\n]", file=file)


@formatter("json-stream")
def format_json_stream(
    directory_entries: Iterable[DirectoryEntry],
    file: TextIO = stdout,
) -> None:
    for inode, type, name in directory_entries:
        print(json_encode(inode, type, name), file=file)



import re
from sys import stdout
from typing import List, Tuple
from unittest.mock import Mock

from pytest import CaptureFixture, MonkeyPatch, mark, raises

import getdents.cli as cli
from getdents.cli import main, parse_args
from getdents.formatters import (
    FORMATTERS,
    Formatter,
    format_csv,
    format_json,
    format_plain,
)


@mark.parametrize(
    ["args", "expected"],
    [
        (["/tmp"], ("/tmp", 32768, format_plain)),
        (["-b", "1234", "x", "-o", "json"], ("x", 1234, format_json)),
        (
            [
                "--buffer-size",
                "9999",
                "--output-format",
                "csv",
                "xxx",
            ],
            ("xxx", 9999, format_csv),
        ),
    ],
)
def test_parse_args(args: List[str], expected: Tuple[str, int, Formatter]) -> None:
    assert parse_args(args, "test") == expected


def test_parse_args_min_buff_size(capsys: CaptureFixture[str]) -> None:
    with raises(SystemExit):
        parse_args(["test", "-b", "0"], "test")

    _, err = capsys.readouterr()

    assert re.search(r"Minimum buffer size is \d+", err) is not None


def test_main(monkeypatch: MonkeyPatch) -> None:
    format_test = Mock()
    getdents = Mock()
    directory_entries = getdents.return_value

    monkeypatch.setitem(FORMATTERS, "test", format_test)
    monkeypatch.setattr(cli, "getdents", getdents)

    assert main(["x", "-o", "test", "-b", "1024"], "test") == 0

    getdents.assert_called_once_with("x", buff_size=1024)
    format_test.assert_called_once_with(directory_entries, stdout)


def test_main_memory_error(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setattr(cli, "getdents", Mock(side_effect=MemoryError))

    assert main(["x"]) == 3


def test_main_file_not_found_error(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setattr(cli, "getdents", Mock(side_effect=FileNotFoundError))

    assert main(["x"]) == 4


def test_main_not_a_directory_error(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setattr(cli, "getdents", Mock(side_effect=NotADirectoryError))

    assert main(["x"]) == 5


def test_main_permission_error(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setattr(cli, "getdents", Mock(side_effect=PermissionError))

    assert main(["x"]) == 6


def test_main_os_error(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setattr(cli, "getdents", Mock(side_effect=OSError))

    assert main(["x"]) == 7



#!/usr/bin/env python
from setuptools import Extension, setup

setup(
    options={"bdist_wheel": {"py_limited_api": "cp38"}},
    ext_modules=[
        Extension(
            "getdents._getdents",
            sources=["getdents/_getdents.c"],
            define_macros=[("Py_LIMITED_API", "0x030800f0")],
            py_limited_api=True,
        ),
    ],
)
